########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.rsm` -- Result Set Management (:xep:`59`)
########################################################

This module provides the :mod:`XSO <aioxmpp.xso>` model for other parts of
:mod:`aioxmpp` and applications which wish to work with :xep:`59` (Result Set
Management).

.. versionadded:: 0.8

.. currentmodule:: aioxmpp.rsm.xso

.. module:: aioxmpp.rsm.xso

XSO
===

.. autoclass:: ResultSetMetadata()

.. autoclass:: After

.. autoclass:: Before

.. autoclass:: First

.. autoclass:: Last
"""
